# Scrapy settings for img_scrapy project
#
# For simplicity, this file contains only settings considered important or
# commonly used. You can find more settings consulting the documentation:
#
#     https://docs.scrapy.org/en/latest/topics/settings.html
#     https://docs.scrapy.org/en/latest/topics/downloader-middleware.html
#     https://docs.scrapy.org/en/latest/topics/spider-middleware.html

BOT_NAME = 'img_scrapy'

SPIDER_MODULES = ['img_scrapy.spiders']
NEWSPIDER_MODULE = 'img_scrapy.spiders'


# Crawl responsibly by identifying yourself (and your website) on the user-agent
#USER_AGENT = 'img_scrapy (+http://www.yourdomain.com)'

# Obey robots.txt rules
ROBOTSTXT_OBEY = True

# Configure maximum concurrent requests performed by Scrapy (default: 16)
#CONCURRENT_REQUESTS = 32

# Configure a delay for requests for the same website (default: 0)
# See https://docs.scrapy.org/en/latest/topics/settings.html#download-delay
# See also autothrottle settings and docs
#DOWNLOAD_DELAY = 3
# The download delay setting will honor only one of:
#CONCURRENT_REQUESTS_PER_DOMAIN = 16
#CONCURRENT_REQUESTS_PER_IP = 16

# Disable cookies (enabled by default)
#COOKIES_ENABLED = False

# Disable Telnet Console (enabled by default)
#TELNETCONSOLE_ENABLED = False

# Override the default request headers:
# DEFAULT_REQUEST_HEADERS = {
#   'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
#   'Accept-Language': 'en',
# }

# Enable or disable spider middlewares
# See https://docs.scrapy.org/en/latest/topics/spider-middleware.html
# SPIDER_MIDDLEWARES = {
#    'img_scrapy.middlewares.ImgScrapySpiderMiddleware': 543,
# }

# Enable or disable downloader middlewares
# See https://docs.scrapy.org/en/latest/topics/downloader-middleware.html
# DOWNLOADER_MIDDLEWARES = {
#    'img_scrapy.middlewares.ImgScrapyDownloaderMiddleware': 543,
# }

# Enable or disable extensions
# See https://docs.scrapy.org/en/latest/topics/extensions.html
# EXTENSIONS = {
#    'scrapy.extensions.telnet.TelnetConsole': None,
# }

# Configure item pipelines
# See https://docs.scrapy.org/en/latest/topics/item-pipeline.html
# ITEM_PIPELINES = {
#    'img_scrapy.pipelines.ImgScrapyPipeline': 300,
# }

# Enable and configure the AutoThrottle extension (disabled by default)
# See https://docs.scrapy.org/en/latest/topics/autothrottle.html
#AUTOTHROTTLE_ENABLED = True
# The initial download delay
#AUTOTHROTTLE_START_DELAY = 5
# The maximum download delay to be set in case of high latencies
#AUTOTHROTTLE_MAX_DELAY = 60
# The average number of requests Scrapy should be sending in parallel to
# each remote server
#AUTOTHROTTLE_TARGET_CONCURRENCY = 1.0
# Enable showing throttling stats for every response received:
AUTOTHROTTLE_DEBUG = True

# Enable and configure HTTP caching (disabled by default)
# See https://docs.scrapy.org/en/latest/topics/downloader-middleware.html#httpcache-middleware-settings
#HTTPCACHE_ENABLED = True
#HTTPCACHE_EXPIRATION_SECS = 0
#HTTPCACHE_DIR = 'httpcache'
#HTTPCACHE_IGNORE_HTTP_CODES = []
#HTTPCACHE_STORAGE = 'scrapy.extensions.httpcache.FilesystemCacheStorage'



from abc import ABC, abstractmethod

from bs4 import BeautifulSoup as bs
from nltk import sent_tokenize
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize


class BaseTextCleaner(ABC):
    """Абстрактный класс клининга текста
    """
    @abstractmethod
    def clean(self, text: str) -> str:
        """Очистка текста перед сантимент-анализом

        Args:
            text (str): Очищаемый текст

        Returns:
            str: Очищенный текст
        """
        raise NotImplementedError


class HtmlTextCleaner(BaseTextCleaner):
    """Очистка текста от html тэгов

    Args:
        BaseTextCleaner (BaseTextCleaner): Абстрактный класс
    """

    def clean(self, text: str) -> str:
        """Очистка текста от тэгов

        Args:
            text (str): Текст

        Returns:
            str: Очищеннный текст
        """
        soup = bs(text)
        return soup.get_text()


class InvisibleSymbolsTextCleaner(BaseTextCleaner):
    """Очистка текста от непечатаемых символов

    Args:
        BaseTextCleaner (BaseTextCleaner): Абстрактный класс
    """

    def clean(self, text: str) -> str:
        """Очистка текста от непечатаемых символов

        Args:
            text (str): Текст

        Returns:
            str: Очищенный текст
        """
        text = text.replace('\t', '')
        text = text.replace('\n\n', '')
        text = text.replace('\n', '')
        return text


class StopWordsTextCleaner(BaseTextCleaner):
    """Очистка текста от стоп слов

    Args:
        BaseTextCleaner (BaseTextCleaner): Абстрактный класс
    """

    def __init__(self, lang: str = 'russian') -> None:
        self.stop_words = stopwords.words('russian')
        self.lang = lang

    def clean(self, text: str) -> str:
        """Очистка текста от стоп слов

        Args:
            text (str): Текст

        Returns:
            str: Очищенный текст
        """

        sentences = sent_tokenize(text, self.lang)
        l_sent = len(sentences)
        for i in range(l_sent):
            tokens = word_tokenize(sentences[i], language=self.lang)
            words = [word for word in tokens if word.isalpha()]
            words = [w for w in words if not w in self.stop_words]
            sentences[i] = " ".join(words)
        return " ".join(sentences)



import re
from typing import List

import requests
from scraping_subsystem.scraper.items import Review
from scraping_subsystem.scraper.spiders.generator_start_urls import \
    BaseGeneratorStartUrl
from scrapy.spiders.sitemap import Spider


class FlampSpider(Spider):
    """Скраппер для сайта flamp.ru

    Args:
        SitemapSpider (SitemapSpider): Класс Краулера

    Raises:
        ValueError: Нет доступа к flamp.js
    """

    name = 'flamp_spider'

    def parse(self, response):
        review = Review()
        review['review_url'] = response.url
        review['author'] = response.xpath(
            "//a[@class='link name t-text t-text--bold']/text()").get()
        review['review_date'] = response.xpath(
            "//span[@class='ugc-date t-text t-text--small']/text()").get()
        review['text_data'] = response.xpath(
            "//p[@class='t-rich-text__p']").get()
        return review


class GeneratorStartUrlFlampSpider(BaseGeneratorStartUrl):
    """Генератор стартовых ссылок для краулера flamp.ru
    """

    def __init__(self, name_company_uri: str) -> None:
        """

        Args:
            name_company_uri (str): Идентификатор ресураса компании
        """
        self.name_company_uri = name_company_uri

    def __get_project_codes(self) -> List[str]:
        """Возвращает доменты для сайта flamp.ru
        <name_domain>.flamp.ru

        Raises:
            ValueError: Нет доступа к flamp.js

        Returns:
            List[str]: Список доменов
        """
        flamp_js_url = 'https://flamp.ru/flamp.js?v=0.1'
        response = requests.get(flamp_js_url)
        if response.status_code == 200:
            project_codes_raw = re.search(
                r'PROJECT_CODES\s*=\s*\[.*\'\],', response.text).group(0)
            project_codes_raw = re.search(
                r'\[.*\]', project_codes_raw).group(0)
            project_codes_raw = project_codes_raw.replace('[', '')
            project_codes_raw = project_codes_raw.replace(']', '')
            project_codes_raw = project_codes_raw.replace('\'', '')
            project_codes = project_codes_raw.split(',')
            return project_codes
        else:
            raise ValueError('Can\'t extract project codes from JS')

    def __get_sitemaps_of_sitemaps_urls(self) -> List[str]:
        """Возвращает карты сайтов доменов с картами сайтов

        Returns:
            List[str]: _description_
        """
        project_codes = self.__get_project_codes()
        sitemaps_of_simaps = [
            f'https://{project_code}.flamp.ru/sitemap.xml' for project_code in project_codes]
        return sitemaps_of_simaps

    def __get_url_from_sitemaps(self, sitemaps_of_sitemaps: List[str]) -> List[str]:
        """Парсит карты сайтов с картами сайтов

        Args:
            sitemaps_of_sitemaps (List[str]): Карты сайтов с картами сайтов

        Returns:
            List[str]: Карты сайтов
        """
        urls = []
        for sitemap in sitemaps_of_sitemaps:
            response = requests.get(sitemap)
            if response.status_code == 200:
                tags = re.findall(r"(<loc>)(.*)(</loc>)", response.text)
                urls.extend([tag[1] for tag in tags])
        return urls

    def get_reviews_start_url(self) -> List[str]:
        """Возвращает стартовые страницы для парсинга

        Returns:
            List[str]: Список ссылок
        """
        sitemaps_of_sitemaps = self.__get_sitemaps_of_sitemaps_urls()
        sitemaps = self.__get_url_from_sitemaps(sitemaps_of_sitemaps)
        started_urls = self.__get_url_from_sitemaps(sitemaps)

        pat = re.compile(f"(.*firm)(\/{self.name_company_uri}.*)(\/otzyv.*)")
        filtered = [url for url in started_urls if pat.match(url)]
        return filtered



# Define here the models for your scraped items
#
# See documentation in:
# https://docs.scrapy.org/en/latest/topics/items.html

import scrapy


class Review(scrapy.Item):
    """Класс контейнер для хранения данных скраппинга

    Args:
        scrapy (_type_): _description_
    """
    review_url = scrapy.Field()
    author = scrapy.Field()
    review_date = scrapy.Field()
    text_data = scrapy.Field()

class Product(scrapy.Item):
    product_url = scrapy.Field()
    price = scrapy.Field()
    title = scrapy.Field()
    img_url = scrapy.Field()



import json
import logging
import tempfile
from typing import Dict

from kafka import KafkaProducer
from scrapy import Spider
from scrapy.crawler import CrawlerProcess
from scrapy.settings import Settings

from scraping_subsystem.kafka_handler.kafka_handler import (TopicHandler,
                                                            result_converter)
from scraping_subsystem.scraper.spiders.flamp_spider import (
    FlampSpider, GeneratorStartUrlFlampSpider)
from scraping_subsystem.sentiment_analyzer.analyzer import SentimentAnalyzer
from scraping_subsystem.sentiment_analyzer.text_cleaner import (
    HtmlTextCleaner, InvisibleSymbolsTextCleaner, StopWordsTextCleaner)


def extract_reviews(spider: Spider, spider_settings: Dict = None,
                    crawl_settings: Dict = None) -> str:
    """С помощью конкретного паука извлекает отзывы компании

    Args:
        spider (Spider): Паук
        spider_settings (Dict): Настройки для класса паука
        crawl_settings (Dict): Настройки для CrawlerProcess

    Returns:
        str: json с отзывами
    """
    with tempfile.TemporaryDirectory() as tempdir:
        result_file_path = f"{tempdir}/reviews.json"
        if spider_settings is None:
            settings = {
                'FEED_FORMAT': "json",
                "FEED_URI": result_file_path}
        else:
            settings = spider_settings
        if crawl_settings is not None:
            crawling_settings = crawl_settings

        crawler_settings = Settings(settings)
        crawler_settings.setmodule(crawler_settings)

        process = CrawlerProcess(settings=crawler_settings)
        process.crawl(spider, **crawling_settings)
        process.start(stop_after_crawl=True)

        with open(result_file_path, 'r', encoding='utf-8') as f:
            result = f.read()
    return result


def calculation_sentiment_marks(reviews: str) -> str:
    """Вычисление тональных оценок

    Args:
        reviews (str): Json с отзывами

    Returns:
        str: _description_
    """
    cleaners = [HtmlTextCleaner(), InvisibleSymbolsTextCleaner(),
                StopWordsTextCleaner()]

    sentiment_analyzer = SentimentAnalyzer(cleaners)
    return sentiment_analyzer.analyze(reviews)


def load_to_kafka(data_reviews: str, key: str,
                  kafka_address: str,
                  name_topic: str = "reviews") -> None:
    """Загружает результаты в Кафку

    Args:
        data_reviews (str): Отзывы с оценками
        key (str): Ключ пакета
        kafka_address (str): адресс Кафки
        name_topic (str, optional): _description_. Defaults to "reviews".
    """
    records = json.loads(data_reviews)

    TopicHandler.create_topic(name_topic, kafka_address)
    producer = KafkaProducer(
        bootstrap_servers=[kafka_address], compression_type='gzip',
        value_serializer=lambda v: json.dumps(
            v, default=result_converter).encode('utf-8')
    )

    for record in records:
        producer.send(topic=name_topic,
                      key=bytes(f"{key}", 'utf-8'), value=record).get(timeout=10)
    producer.flush()
    logging.info(
        f"Loaded {len(records)} unmatched results {key} to Apache Kafka")


if __name__ == '__main__':
    logging.basicConfig(
        level=logging.NOTSET,
        filename="crawler_logs.log",
        format="%(asctime)s - %(module)s - %(levelname)s - %(funcName)s: %(lineno)d - %(message)s",
        datefmt='%H:%M:%S',
    )

    generator_start_urls = GeneratorStartUrlFlampSpider(
        'kfc_set_restoranov_bystrogo_obsluzhivaniya')
    start_urls = generator_start_urls.get_reviews_start_url()
    reviews = extract_reviews(FlampSpider, crawl_settings={
                              'start_urls': start_urls[:100]})
    reviews = calculation_sentiment_marks(reviews)
    load_to_kafka(reviews, 'flamp.ru/kfc', 'localhost:9092')



# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html


# useful for handling different item types with a single interface
from itemadapter import ItemAdapter


class ImgScrapyPipeline:
    def process_item(self, item, spider):
        return item



#! /usr/bin/python3.8
import json
import logging
from typing import Dict, List

import fasttext
from dostoevsky.models import FastTextSocialNetworkModel
from dostoevsky.tokenization import RegexTokenizer
from nltk import download as nltk_download
from scraping_subsystem.sentiment_analyzer.text_cleaner import BaseTextCleaner
from tqdm import tqdm


class SentimentAnalyzer:
    """Класс сантимент-анализа отзывов
    """

    def __init__(self, text_cleaners: List[BaseTextCleaner] = None) -> None:
        nltk_download('all')
        nltk_download('stopwords')
        logging.info('Downloaded nltk data')
        self.tokenizer = RegexTokenizer()
        logging.info("set tokenizer")
        fasttext.FastText.eprint = lambda x: None

        logging.info("Set sropwords")
        if text_cleaners is not None:
            self.text_cleaners = text_cleaners
            logging.info("Set text cleaners")
        else:
            self.text_cleaners = []
            logging.info("Set text cleaners are empty")

    def __clean_text(self, text: str) -> str:
        """Очистка текста с помощью текст клинеров

        Args:
            text (str): Текст

        Returns:
            str: Очищенный текст
        """
        for cleaner in self.text_cleaners:
            text = cleaner.clean(text)
        return text

    def __prepare_sentiment_list(self, sentiment_list: List[Dict]) -> Dict[str, float]:
        """

        Args:
            sentiment_values (_type_): Список словарей с тональными оценками

        Returns:
            Dict[str,float]: Тональные оценки
        """

        prepared_dict = {}

        for sentiment in sentiment_list:
            neutral = sentiment.get('neutral')
            negative = sentiment.get('negative')
            positive = sentiment.get('positive')
            if neutral is None:
                prepared_dict['neutral'] = 0
            else:
                prepared_dict['neutral'] = sentiment['neutral']
            if negative is None:
                prepared_dict['negative'] = 0
            else:
                prepared_dict['negative'] = sentiment['negative']
            if positive is None:
                prepared_dict['positive'] = 0
            else:
                prepared_dict['positive'] = sentiment['positive']
        return prepared_dict

    def analyze(self, reviews: str) -> str:
        """Сантимент-анализ отзывов

        Args:
            reviws (str): Json с отзывами

        Returns:
            str: Отзывы с тональными оценками json
        """
        reviews_list = json.loads(reviews)
        result_reviews = []
        for review in tqdm(reviews_list):
            model = FastTextSocialNetworkModel(
                tokenizer=self.tokenizer, lemmatize=True)
            text = self.__clean_text(review['text_data'])

            sentiment_list = model.predict([text], k=2)
            sentiment_marks = self.__prepare_sentiment_list(sentiment_list)
            evaluated_review = {**review, **sentiment_marks}
            result_reviews.append(evaluated_review)

        return json.dumps(result_reviews)



import scrapy

from scraper.items import Product

class TestSpider(scrapy.Spider):
    name = 'ecom_spider'
    allowed_domains = ['clever-lichterman-044f16.netlify.com']
    start_urls = [
        'https://clever-lichterman-044f16.netlify.com/products/taba-cream.1/']

    def parse(self, response):
        item = Product()
        item['product_url'] = response.url
        item['price'] = response.xpath(
            "//div[@class='my-4']/span/text()").get()
        item['title'] = response.xpath('//section[1]//h2/text()').get()
        item['img_url'] = response.xpath(
            "//div[@class='product-slider']//img/@src").get(0)
        return item



import logging
from datetime import date, datetime
from typing import Dict
from uuid import UUID

from kafka.admin import KafkaAdminClient, NewPartitions, NewTopic
from kafka.errors import InvalidPartitionsError, TopicAlreadyExistsError


class TopicHandler:
    """Класс для работы с топиками
    """
    @staticmethod
    def create_topic(name_topic: str,
                     kafka_address: str) -> None:
        """Создаёт топик в Кафка если он ещё не был создан

        Args:
            name_topic (str): Имя топика
        """
        conf = {'bootstrap_servers': kafka_address}
        client = KafkaAdminClient(**conf)
        topics = []
        topics.append(NewTopic(name_topic, 1, 1))
        try:
            client.create_topics(topics)
            logging.info(f"Successfull creating topic {name_topic}")

        except TopicAlreadyExistsError as topic_already_exists_error:
            logging.error(f"{str(topic_already_exists_error)}")
        else:
            try:
                topic_partitions = {}
                topic_partitions[name_topic] = NewPartitions(total_count=3)
                client.create_partitions(topic_partitions)
                logging.info(f"Successfull creating 3 partitions")
            except InvalidPartitionsError as invalid_partitions_error:
                logging.error(f"{invalid_partitions_error}")


def result_converter(obj: Dict) -> str:
    """Json converter для дат
    Raises:
        TypeError: Object not datetime

    Returns:
        str: _description_
    """
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    if isinstance(obj, UUID):
        return obj.hex
    raise TypeError(f"{type(obj)} not datetime")



from abc import ABC, abstractmethod
from typing import List


class BaseGeneratorStartUrl(ABC):
    """Генератор стартовых ссылок
    """

    @abstractmethod
    def get_reviews_start_url(self) -> List[str]:
        """Генерирует стартовые ссылки

        Returns:
            List[str]: список ссылок
        """
        raise NotImplementedError
